@@ -14,7 +14,6 @@ with 'DR::Tnt::Role::Logging';
 use Scalar::Util;
 use feature 'state';
 
-
 use constant SPACE_space        => 281;     # _vspace
 use constant SPACE_index        => 289;     # _vindex
 use constant ER_TNT_PERMISSIONS => 0x8037;
@@ -23,7 +22,13 @@ use constant ER_TNT_SCHEMA      => 0x806D;
 use Mouse::Util::TypeConstraints;
 
     enum DriverType     => [ 'sync', 'async' ];
-    enum FullCbState    => [ 'init', 'connecting', 'schema', 'ready', 'pause' ];
+    enum FullCbState    => [
+        'init',
+        'connecting',
+        'schema',
+        'ready',
+        'pause',
+    ];
 
 no Mouse::Util::TypeConstraints;
 
@@ -53,11 +58,16 @@ has state =>
     writer      => '_set_state',
     trigger     => sub {
         my ($self, undef, $old_state) = @_;
+        $self->_state_changed($self->_now);
 
         $self->_reconnector->event($self->state, $old_state);
         $self->_log(info => 'Connector is in state: %s',  $self->state);
     };
 ;
+
+has _state_changed  => is => 'rw', isa => 'Maybe[Num]';
+
+
 has last_schema =>
     is      => 'ro',
     isa     => 'Int',
@@ -87,76 +97,6 @@ has _reconnector    =>
 ;
 
 
-sub restart {
-    my ($self, $cb) = @_;
-
-
-    $cb ||= sub {  };
-    $self->_log(info => 'Starting connection to %s:%s (driver: %s)',
-        $self->host, $self->port, $self->driver);
-
-    goto $self->state;
-    
-    init:
-    connecting:
-    schema:
-    pause:
-    ready:
-        $self->_set_state('connecting');
-        $self->_reconnector->ll->connect(sub {
-            my ($state, $message) = @_;
-            unless ($state eq 'OK') {
-                $self->_set_last_error([ $state, $message ]);
-                $self->_set_state('pause');
-                $cb->($state => $message);
-                return;
-            }
-
-            $self->_reconnector->ll->handshake(sub {
-                my ($state, $message) = @_;
-                unless ($state eq 'OK') {
-                    $self->_set_last_error([ $state, $message ]);
-                    $self->_set_state('pause');
-                    $cb->($state => $message);
-                    return;
-                }
-
-                unless ($self->user and $self->password) {
-                    return $self->_preeval_lua($cb);
-                }
-
-                $self->_reconnector->ll->send_request(auth => undef, sub {
-                    my ($state, $message, $sync) = @_;
-                    unless ($state eq 'OK') {
-                        $self->_set_last_error([ $state, $message ]);
-                        $self->_set_state('pause');
-                        $cb->($state => $message);
-                        return;
-                    }
-
-                    $self->_reconnector->ll->wait_response($sync, sub {
-                        my ($state, $message, $resp) = @_;
-                        unless ($state eq 'OK') {
-                            $self->_set_last_error([ $state, $message ]);
-                            $self->_set_state('pause');
-                            $cb->($state => $message);
-                            return;
-                        }
-
-                        unless ($resp->{CODE} == 0) {
-                            $self->_set_last_error([ ER_BROKEN_PASSWORD =>
-                                $resp->{ERROR} // 'Wrong password']
-                            );
-                            $self->_set_state('pause');
-                            $cb->(@{ $self->last_error });
-                            return;
-                        }
-                        $self->_preeval_lua($cb);
-                    });
-                });
-            });
-        });
-}
 
 has _unsent_lua     => is => 'rw', isa => 'ArrayRef', default => sub {[]};
 
@@ -184,7 +124,7 @@ sub _preeval_unsent_lua {
 
     my $lua = shift @{ $self->_unsent_lua };
 
-    $self->_log(debug => 'Eval "%s" after connection', $lua); 
+    $self->_log(debug => 'Eval "%s" after connection', $lua);
 
     if (open my $fh, '<:raw', $lua) {
         local $/;
@@ -264,7 +204,7 @@ sub _invalid_schema {
                     $cb->($state => $message);
                     return;
                 }
-                
+
 
                 # have no permissions
                 if ($resp->{CODE} == ER_TNT_PERMISSIONS) {
@@ -281,7 +221,7 @@ sub _invalid_schema {
 
                 $self->_log(debug => 'Loading indexes');
                 $self->_reconnector->ll->send_request(select =>
-                    $resp->{SCHEMA_ID}, SPACE_index, 0, [], undef, undef, 'ALL', sub { 
+                    $resp->{SCHEMA_ID}, SPACE_index, 0, [], undef, undef, 'ALL', sub {
 
                     my ($state, $message, $sync) = @_;
                     unless ($state eq 'OK') {
@@ -300,14 +240,14 @@ sub _invalid_schema {
                             $cb->($state => $message);
                             return;
                         }
-                    
+
                         if ($resp->{CODE} == ER_TNT_PERMISSIONS) {
                             $self->_indexes([]);
                         } elsif ($resp->{CODE} == ER_TNT_SCHEMA) {
                             # collision again!
                             $self->_invalid_schema($cb);
                             return;
-                        
+
                         } elsif ($resp->{CODE}) {
                             $self->_set_last_error([ ER_REQUEST =>
                                 'Can not load tarantool schema', $resp->{CODE} ]);
@@ -322,10 +262,8 @@ sub _invalid_schema {
                         $self->_set_schema($resp->{SCHEMA_ID});
                         $self->_set_state('ready');
 
-                        my $list = $self->_wait_ready;
-                        $self->_wait_ready([]);
                         $cb->('OK', 'Connected, schema loaded');
-                        $self->request(@$_) for @$list;
+                        $self->request;
                     });
                 });
             });
@@ -350,7 +288,7 @@ sub _set_schema {
         for (@{ $self->_indexes }) {
             next unless $_->[0] == $space->{id};
 
-            $space->{indexes}{ $_->[2] } = 
+            $space->{indexes}{ $_->[2] } =
             $space->{indexes}{ $_->[1] } = {
                 id      => $_->[1],
                 name    => $_->[2],
@@ -362,7 +300,7 @@ sub _set_schema {
             }
         }
     }
-     
+
     $self->_set_last_schema($schema_id);
     $self->_sch(\%sch);
     $self->_indexes([]);
@@ -408,150 +346,326 @@ sub _tuples {
     $cb->(OK => 'Response received', $res);
 }
 
-sub request {
-    my $cb = pop;
-    my ($self, $name, @args) = @_;
+sub restart {
+    my ($self, $cbc) = @_;
+
+    $cbc ||= sub {  };
+
+    $self->_log(info => 'Starting connection to %s:%s (driver: %s)',
+        $self->host, $self->port, $self->driver);
 
-    my $request = [ $name, @args, $cb ];
-   
-    # all states waits ready
     goto $self->state;
 
     init:
-        $self->_log(info => 'Defer request "%s" until first connected', $name);
-        push @{ $self->_wait_ready } => $request;
-        
+    connecting:
+    schema:
+    pause:
+    ready:
+        $self->_set_state('connecting');
+        $self->_reconnector->ll->connect(sub {
+            my ($state, $message) = @_;
+            unless ($state eq 'OK') {
+                $self->_set_last_error([ $state, $message ]);
+                $self->_set_state('pause');
+                $cbc->(@{ $self->last_error });
+                return;
+            }
+
+            $self->_reconnector->ll->handshake(sub {
+                my ($state, $message) = @_;
+                unless ($state eq 'OK') {
+                    $self->_set_last_error([ $state, $message ]);
+                    $self->_set_state('pause');
+                    $cbc->(@{ $self->last_error });
+                    return;
+                }
+
+                unless ($self->user and $self->password) {
+                    return $self->_preeval_lua($cbc);
+                }
+
+                $self->_reconnector->ll->send_request(auth => undef, sub {
+                    my ($state, $message, $sync) = @_;
+                    unless ($state eq 'OK') {
+                        $self->_set_last_error([ $state, $message ]);
+                        $self->_set_state('pause');
+                        $cbc->(@{ $self->last_error });
+                        return;
+                    }
+
+                    $self->_reconnector->ll->wait_response($sync, sub {
+                        my ($state, $message, $resp) = @_;
+                        unless ($state eq 'OK') {
+                            $self->_set_last_error([ $state, $message ]);
+                            $self->_set_state('pause');
+                            $cbc->(@{ $self->last_error });
+                            return;
+                        }
+
+                        unless ($resp->{CODE} == 0) {
+                            $self->_log(warning => 'Can not auth: Wrong login or password');
+                            $self->_set_last_error([ ER_BROKEN_PASSWORD =>
+                                $resp->{ERROR} // 'Wrong password']
+                            );
+                            $self->_set_state('pause');
+                            $cbc->(@{ $self->last_error });
+                            return;
+                        }
+                        $self->_preeval_lua($cbc);
+                    });
+                });
+            });
+        });
+}
+
+sub request {
+    my $self = shift;
+
+    if (@_) {
+
+        unless ('CODE' eq ref $_[-1]) {
+            croak 'usage: $connector->request(..., $CALLBACK)';
+        }
+        state $check = {
+            get         => sub {
+                croak 'usage: $connector->get(space, index, key)'
+                    unless @_ == 5;
+            },
+            select      => sub {
+                croak 'usage: $connector->select(space, index, key[, limit, offset, iterator])'
+                    unless @_ >= 5 and @_ <= 8;
+            },
+            update      => sub { },
+            insert      => sub {
+                croak 'usage: $connector->insert(space, tuple)'
+                    unless @_ == 4 and 'ARRAY' eq ref $_[2];
+            },
+            replace     => sub {
+                croak 'usage: $connector->replace(space, tuple)'
+                    unless @_ == 4 and 'ARRAY' eq ref $_[2];
+            },
+            delete      => sub {
+                croak 'usage: $connector->delete(space, key)'
+                    unless @_ == 4;
+            },
+            call_lua    => sub {
+                croak 'usage: $connector->call_lua(name[, args])'
+                    unless @_ >= 3;
+            },
+            eval_lua    => sub {
+                croak 'usage: $connector->eval_lua(code[, args])'
+                    unless @_ >= 3;
+            },
+            ping        => sub {
+            },
+            auth        => sub {
+                croak 'usage: $connector->auth([user, password])'
+                    unless @_ == 3 or @_ == 5;
+            },
+        };
+
+        unless (exists $check->{ $_[0] // 'undef' }) {
+            croak 'unknown request method: ' . $_[0] // 'undef';
+        }
+
+        $check->{ $_[0] }(@_);
+
+        push @{ $self->_wait_ready } => \@_;
+    }
+
+    restart:
+        goto $self->state;
+
+
+    init:
         $self->_log(info => 'Autoconnect before first request');
-        $self->restart;
+
+    reinit:
+        $self->restart(sub {
+            return if $self->state eq 'ready';
+            unless (defined $self->reconnect_interval) {
+                my $list = $self->_wait_ready;
+                $self->_wait_ready([]);
+                for (@$list) {
+                    $_->[-1](@{ $self->last_error });
+                }
+                return;
+            }
+        });
+
+        if ($self->driver eq 'sync') {
+            goto ready if $self->state eq 'ready';
+            goto sync_redo_check;
+        }
+
         return;
+    
+    sync_redo_check:
+        goto no_reconnect_errors unless defined $self->reconnect_interval;
+        Time::HiRes::sleep($self->reconnect_interval);
+        goto reinit;
 
     schema:
     connecting:
-        $self->_log(info => 'Defer request "%s" until schema loaded', $name);
-        push @{ $self->_wait_ready } => $request;
         return;
 
     pause:
-        unless (defined $self->reconnect_interval) {
-            $cb->(@{ $self->last_error });
-            return;
+        if ($self->driver eq 'sync') {
+            goto no_reconnect_errors unless defined $self->reconnect_interval;
+            my $pause = $self->reconnect_interval -
+                ($self->_now - $self->_state_changed);
+            Time::HiRes::sleep($pause) if $pause > 0;
+            goto reinit;
         }
-
-        $self->_log(info => 'Defer request "%s" until reconnected', $name);
-        push @{ $self->_wait_ready } => $request;
-        $self->_reconnector->check_pause;
+        goto no_reconnect_errors unless defined $self->reconnect_interval;
         return;
+    no_reconnect_errors: {
+            my $list = $self->_wait_ready;
+            $self->_wait_ready([]);
+            for (@$list) {
+                $_->[-1](@{ $self->last_error });
+            }
+            return
+    }
 
     ready:
+        while (my $request = shift @{ $self->_wait_ready }) {
 
-    croak 'Usage $tnt->request(method => args ... sub { .. })'
-        unless 'CODE' eq ref $cb;
+            unless ($self->state eq 'ready') {
+                unshift @{ $self->_wait_ready  } => $request;
+                goto restart;
+            };
+
+            my @args = @$request;
+            my $name = shift @args;
+            my $cb = pop @args;
+
+            my ($space, $index);
+            state $space_pos = {
+                select      => 'index',
+                update      => 'normal',
+                insert      => 'normal',
+                replace     => 'normal',
+                delete      => 'normal',
+                call_lua    => 'mayberef',
+                eval_lua    => 'mayberef',
+                ping        => 'none',
+                auth        => 'none',
+            };
+
+            croak "unknown method $name" unless exists $space_pos->{$name};
+
+            goto $space_pos->{$name};
+
+            index:
+                $space = $args[0];
+                unless (exists $self->_sch->{ $space }) {
+                    $self->_set_last_error([ER_NOSPACE => "Space $space not found"]);
+                    $cb->(@{ $self->last_error });
+                    next;
+                }
+                $args[0] = $self->_sch->{ $space }{id};
+
+                $index = $args[1];
+                unless (exists $self->_sch->{ $space }{indexes}{ $index }) {
+                    $self->_set_last_error(
+                        [ER_NOINDEX => "Index space[$space].$index not found"]
+                    );
+                    $cb->(@{ $self->last_error });
+                    next;
+                }
+
+                $index = $args[1] = $self->_sch->{$space}{indexes}{ $index }{id};
+                goto do_request;
+
+            normal:
+                $space = $args[0];
+                unless (exists $self->_sch->{ $space }) {
+                    $self->_set_last_error([ER_NOSPACE => "Space $space not found"]);
+                    $cb->(@{ $self->last_error });
+                    next;
+                }
+                $space = $args[0] = $self->_sch->{ $space }{id};
+                goto do_request;
+
+            mayberef:
+                if ('ARRAY' eq ref $args[0]) {
+                    ($args[0], $space) = @{ $args[0] };
+                }
+                goto do_request unless defined $space;
+                unless (exists $self->_sch->{ $space }) {
+                    $self->_set_last_error([ER_NOSPACE => "Space $space not found"]);
+                    $cb->(@{ $self->last_error });
+                    next;
+                }
+                $space = $self->_sch->{ $space }{id};
 
 
-    my ($space, $index);
-    state $space_pos = {
-        select      => 'index',
-        update      => 'normal',
-        insert      => 'normal',
-        replace     => 'normal',
-        delete      => 'normal',
-        call_lua    => 'mayberef',
-        eval_lua    => 'mayberef',
-        ping        => 'none',
-        auth        => 'none',
-    };
+            none:
 
-    croak "unknown method $name" unless exists $space_pos->{$name};
+            do_request:
 
-    goto $space_pos->{$name};
+                $self->_reconnector->ll->send_request($name, $self->last_schema,
+                    @args, sub {
+                        my ($state, $message, $sync) = @_;
+                        unless ($state eq 'OK') {
+                            $self->_set_last_error([ $state => $message ]);
+                            $self->_set_state('pause');
+                            $cb->(@{ $self->last_error });
+                            return;
+                        }
 
-    index:
-        $space = $args[0];
-        unless (exists $self->_sch->{ $space }) {
-            $self->_set_last_error([ER_NOSPACE => "Space $space not found"]);
-            $cb->(@{ $self->last_error });
-            return;
-        }
-        $args[0] = $self->_sch->{ $space }{id};
-        
-        $index = $args[1];
-        unless (exists $self->_sch->{ $space }{indexes}{ $index }) {
-            $self->_set_last_error(
-                [ER_NOINDEX => "Index space[$space].$index not found"]
-            );
-            $cb->(@{ $self->last_error });
-            return;
-        }
+                        $self->_reconnector->ll->wait_response($sync, sub {
+                            my ($state, $message, $resp) = @_;
+                            unless ($state eq 'OK') {
+                                $self->_set_last_error([ $state => $message ]);
+                                $self->_set_state('pause');
+                                $cb->(@{ $self->last_error });
+                                return;
+                            }
 
-        $index = $args[1] = $self->_sch->{$space}{indexes}{ $index }{id};
-        goto do_request;
+                            # schema collision
+                            if ($resp->{CODE} == ER_TNT_SCHEMA) {
+                                $self->_log(warning => 'Detected schema collision');
+                                $self->_log(info => 'Defer request "%s" until schema loaded', $name);
+                                unshift @{ $self->_wait_ready } => $request;
+                                $self->_invalid_schema(sub {}) if $self->state eq 'ready';
+                                return;
+                            }
 
-    normal:
-        $space = $args[0];
-        unless (exists $self->_sch->{ $space }) {
-            $self->_set_last_error([ER_NOSPACE => "Space $space not found"]);
-            $cb->(@{ $self->last_error });
-            return;
-        }
-        $space = $args[0] = $self->_sch->{ $space }{id};
-        goto do_request;
+                            unless ($resp->{CODE} == 0) {
+                                $self->_set_last_error(
+                                    [ ER_REQUEST => $resp->{ERROR}, $resp->{CODE} ]
+                                );
+                                $cb->(@{ $self->last_error });
+                                return;
+                            }
 
-    mayberef:
-        if ('ARRAY' eq ref $args[0]) {
-            ($args[0], $space) = @{ $args[0] };
-        }
-        goto do_request unless defined $space;
-        unless (exists $self->_sch->{ $space }) {
-            $self->_set_last_error([ER_NOSPACE => "Space $space not found"]);
-            $cb->(@{ $self->last_error });
-            return;
-        }
-        $space = $self->_sch->{ $space }{id};
+                            if ($resp->{SCHEMA_ID} != $self->last_schema) {
+                                $self->_log(info => 'request was changed schema id');
+                            }
+                            $self->_set_last_error(undef);
+                            $self->_tuples($resp, $space, $cb);
+                        });
+                    });
+    }
+    return;
 
-
-    none:
-
-    do_request:
-
-    $self->_reconnector->ll->send_request($name, $self->last_schema, @args, sub {
-        my ($state, $message, $sync) = @_;
-        unless ($state eq 'OK') {
-            $self->_set_last_error([ $state => $message ]);
-            $self->_set_state('pause');
-            $cb->(@{ $self->last_error });
-            return;
-        }
-
-        $self->_reconnector->ll->wait_response($sync, sub {
-            my ($state, $message, $resp) = @_;
-            unless ($state eq 'OK') {
-                $self->_set_last_error([ $state => $message ]);
-                $self->_set_state('pause');
-                $cb->(@{ $self->last_error });
-                return;
-            }
-
-            # schema collision
-            if ($resp->{CODE} == ER_TNT_SCHEMA) {
-                $self->_log(error => 'Detected schema collision');
-                $self->_log(info => 'Defer request "%s" until schema loaded', $name);
-                push @{ $self->_wait_ready } => $request;
-                $self->_invalid_schema(sub {}) if $self->state eq 'ready';
-                return;
-            }
-
-            unless ($resp->{CODE} == 0) {
-                $self->_set_last_error(
-                    [ ER_REQUEST => $resp->{ERROR}, $resp->{CODE} ]
-                );
-                $cb->(@{ $self->last_error });
-                return;
-            }
-            $self->_set_last_error(undef);
-            $self->_tuples($resp, $space, $cb);
-        });
-    });
 }
 
+sub _now {
+    my ($self) = @_;
+    return Time::HiRes::time() if $self->driver eq 'sync';
+    return AnyEvent::now();
+}
+
+sub BUILD {
+    my ($self) = @_;
+    if ($self->driver eq 'sync') {
+        require Time::HiRes;
+    } else {
+        require AnyEvent;
+    }
+}
 __PACKAGE__->meta->make_immutable;
 
